# External Libraries
from unicodedata import east_asian_width
from wcwidth import wcwidth
from dataclasses import dataclass
import time, re, logging

from typing import Iterable, NamedTuple, Optional, Union, Tuple
from prompt_toolkit import ANSI
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.formatted_text import to_formatted_text, fragment_list_to_text
from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.processors import Processor, Transformation
from prompt_toolkit.application.current import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.data_structures import Point
from prompt_toolkit.layout.controls import UIContent, FormattedTextControl
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.selection import SelectionType
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.utils import Event

from prompt_toolkit.filters import (
    FilterOrBool,
)
from prompt_toolkit.formatted_text import (
    StyleAndTextTuples,
    to_formatted_text,
)
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding.key_bindings import KeyBindingsBase
from prompt_toolkit.layout.containers import (
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import (
    
    FormattedTextControl,
)
from prompt_toolkit.layout.processors import (
    Processor,
    TransformationInput,
    Transformation
)
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import get_cwidth
from prompt_toolkit.widgets import Button, MenuContainer, MenuItem
from prompt_toolkit.widgets.base import Border

from prompt_toolkit.layout.screen import _CHAR_CACHE, Screen, WritePosition
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.formatted_text.utils import (
    fragment_list_to_text,
    fragment_list_width,
)

from .settings import Settings

class VSplitWindow(Window):
    "修改的分块窗口，向上翻页时，下半部保持最后数据不变"
 
    def _copy_body(
        self,
        ui_content: UIContent,
        new_screen: Screen,
        write_position: WritePosition,
        move_x: int,
        width: int,
        vertical_scroll: int = 0,
        horizontal_scroll: int = 0,
        wrap_lines: bool = False,
        highlight_lines: bool = False,
        vertical_scroll_2: int = 0,
        always_hide_cursor: bool = False,
        has_focus: bool = False,
        align: WindowAlign = WindowAlign.LEFT,
        get_line_prefix = None,
        isNotMargin = True,
    ):
        """
        Copy the UIContent into the output screen.
        Return (visible_line_to_row_col, rowcol_to_yx) tuple.

        :param get_line_prefix: None or a callable that takes a line number
            (int) and a wrap_count (int) and returns formatted text.
        """
        xpos = write_position.xpos + move_x
        ypos = write_position.ypos
        line_count = ui_content.line_count
        new_buffer = new_screen.data_buffer
        empty_char = _CHAR_CACHE["", ""]

        # Map visible line number to (row, col) of input.
        # 'col' will always be zero if line wrapping is off.
        visible_line_to_row_col: dict[int, tuple[int, int]] = {}

        # Maps (row, col) from the input to (y, x) screen coordinates.
        rowcol_to_yx: dict[tuple[int, int], tuple[int, int]] = {}

        def copy_line(
            line: StyleAndTextTuples,
            lineno: int,
            x: int,
            y: int,
            is_input: bool = False,
        ):
            """
            Copy over a single line to the output screen. This can wrap over
            multiple lines in the output. It will call the prefix (prompt)
            function before every line.
            """
            if is_input:
                current_rowcol_to_yx = rowcol_to_yx
            else:
                current_rowcol_to_yx = {}  # Throwaway dictionary.

            # Draw line prefix.
            if is_input and get_line_prefix:
                prompt = to_formatted_text(get_line_prefix(lineno, 0))
                x, y = copy_line(prompt, lineno, x, y, is_input=False)

            # Scroll horizontally.
            skipped = 0  # Characters skipped because of horizontal scrolling.
            if horizontal_scroll and is_input:
                h_scroll = horizontal_scroll
                line = explode_text_fragments(line)
                while h_scroll > 0 and line:
                    h_scroll -= get_cwidth(line[0][1])
                    skipped += 1
                    del line[:1]  # Remove first character.

                x -= h_scroll  # When scrolling over double width character,
                # this can end up being negative.

            # Align this line. (Note that this doesn't work well when we use
            # get_line_prefix and that function returns variable width prefixes.)
            if align == WindowAlign.CENTER:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += (width - line_width) // 2
            elif align == WindowAlign.RIGHT:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += width - line_width

            col = 0
            wrap_count = 0
            for style, text, *_ in line:
                new_buffer_row = new_buffer[y + ypos]

                # Remember raw VT escape sequences. (E.g. FinalTerm's
                # escape sequences.)
                if "[ZeroWidthEscape]" in style:
                    new_screen.zero_width_escapes[y + ypos][x + xpos] += text
                    continue

                for c in text:
                    char = _CHAR_CACHE[c, style]
                    char_width = char.width

                    # Wrap when the line width is exceeded.
                    if wrap_lines and x + char_width > width:
                        visible_line_to_row_col[y + 1] = (
                            lineno,
                            visible_line_to_row_col[y][1] + x,
                        )
                        y += 1
                        wrap_count += 1
                        x = 0

                        # Insert line prefix (continuation prompt).
                        if is_input and get_line_prefix:
                            prompt = to_formatted_text(
                                get_line_prefix(lineno, wrap_count)
                            )
                            x, y = copy_line(prompt, lineno, x, y, is_input=False)

                        new_buffer_row = new_buffer[y + ypos]

                        if y >= write_position.height:
                            return x, y  # Break out of all for loops.

                    # Set character in screen and shift 'x'.
                    if x >= 0 and y >= 0 and x < width:
                        new_buffer_row[x + xpos] = char

                        # When we print a multi width character, make sure
                        # to erase the neighbours positions in the screen.
                        # (The empty string if different from everything,
                        # so next redraw this cell will repaint anyway.)
                        if char_width > 1:
                            for i in range(1, char_width):
                                new_buffer_row[x + xpos + i] = empty_char

                        # If this is a zero width characters, then it's
                        # probably part of a decomposed unicode character.
                        # See: https://en.wikipedia.org/wiki/Unicode_equivalence
                        # Merge it in the previous cell.
                        elif char_width == 0:
                            # Handle all character widths. If the previous
                            # character is a multiwidth character, then
                            # merge it two positions back.
                            for pw in [2, 1]:  # Previous character width.
                                if (
                                    x - pw >= 0
                                    and new_buffer_row[x + xpos - pw].width == pw
                                ):
                                    prev_char = new_buffer_row[x + xpos - pw]
                                    char2 = _CHAR_CACHE[
                                        prev_char.char + c, prev_char.style
                                    ]
                                    new_buffer_row[x + xpos - pw] = char2

                        # Keep track of write position for each character.
                        current_rowcol_to_yx[lineno, col + skipped] = (
                            y + ypos,
                            x + xpos,
                        )

                    col += 1
                    x += char_width
            return x, y

        # Copy content.
        def copy() -> int:
            y = -vertical_scroll_2        
            lineno = vertical_scroll
            
            total = write_position.height
            upper = (total - 1) // 2
            below = total - upper - 1
            
            #if isNotMargin:
            if isinstance(self.content, SessionBufferControl):
                b = self.content.buffer
                if not b:
                    return y
                    
                line_count = b.lineCount
                start_lineno = b.start_lineno
                if start_lineno < 0:
                    # no split window
                    if line_count < total:
                        # 内容行数小于屏幕行数
                        lineno = 0

                        while y < total and lineno < line_count:
                            # Take the next line and copy it in the real screen.
                            line = ui_content.get_line(lineno)
                            visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                            x = 0
                            x, y = copy_line(line, lineno, x, y, is_input=True)
                            lineno += 1
                            y += 1

                    else:
                        # 若内容行数大于屏幕行数，则倒序复制，确保即使有自动折行时，最后一行也保持在屏幕最底部

                        y = total
                        lineno = line_count

                        while y >= 0 and lineno >= 0:
                            lineno -= 1
                            # Take the next line and copy it in the real screen.
                            display_lines = ui_content.get_height_for_line(lineno, width, None)
                            y -= display_lines
                            line = ui_content.get_line(lineno)
                            visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                            copy_line(line, lineno, 0, y, is_input=True)

                    
                else:
                    # 有split window
                    # 复制上半部分，正序复制，确保即使有自动折行时，第一行也保持在屏幕最顶部
                    lineno = start_lineno
                    while y < upper and lineno < line_count:
                        line = ui_content.get_line(lineno)
                        visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                        x = 0
                        x, y = copy_line(line, lineno, x, y, is_input=True)
                        lineno += 1
                        y += 1

                    # x = 0
                    # x, y = copy_line([("","-"*width)], lineno, x, y, is_input=False)
                    # y += 1

                    # 复制下半部分，倒序复制，确保即使有自动折行时，最后一行也保持在屏幕最底部
                    y = total
                    lineno = line_count

                    while y >= below and lineno >= 0:
                        lineno -= 1
                        # Take the next line and copy it in the real screen.
                        display_lines = ui_content.get_height_for_line(lineno, width, None)
                        y -= display_lines
                        line = ui_content.get_line(lineno)
                        visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                        copy_line(line, lineno, 0, y, is_input=True)

                    # 最后复制分割线，若上下有由于折行额外占用的内容，都用分割线给覆盖掉
                    copy_line([("","-"*width)], -1, 0, upper + 1, is_input=False)
                    
            return y

        copy()

        def cursor_pos_to_screen_pos(row: int, col: int) -> Point:
            "Translate row/col from UIContent to real Screen coordinates."
            try:
                y, x = rowcol_to_yx[row, col]
            except KeyError:
                # Normally this should never happen. (It is a bug, if it happens.)
                # But to be sure, return (0, 0)
                return Point(x=0, y=0)

                # raise ValueError(
                #     'Invalid position. row=%r col=%r, vertical_scroll=%r, '
                #     'horizontal_scroll=%r, height=%r' %
                #     (row, col, vertical_scroll, horizontal_scroll, write_position.height))
            else:
                return Point(x=x, y=y)

        # Set cursor and menu positions.
        if ui_content.cursor_position:
            screen_cursor_position = cursor_pos_to_screen_pos(
                ui_content.cursor_position.y, ui_content.cursor_position.x
            )

            if has_focus:
                new_screen.set_cursor_position(self, screen_cursor_position)

                if always_hide_cursor:
                    new_screen.show_cursor = False
                else:
                    new_screen.show_cursor = ui_content.show_cursor

                self._highlight_digraph(new_screen)

            if highlight_lines:
                self._highlight_cursorlines(
                    new_screen,
                    screen_cursor_position,
                    xpos,
                    ypos,
                    width,
                    write_position.height,
                )

        # Draw input characters from the input processor queue.
        if has_focus and ui_content.cursor_position:
            self._show_key_processor_key_buffer(new_screen)

        # Set menu position.
        if ui_content.menu_position:
            new_screen.set_menu_position(
                self,
                cursor_pos_to_screen_pos(
                    ui_content.menu_position.y, ui_content.menu_position.x
                ),
            )

        # Update output screen height.
        new_screen.height = max(new_screen.height, ypos + write_position.height)

        return visible_line_to_row_col, rowcol_to_yx

    def _scroll_down(self) -> None:
        "向下滚屏，处理屏幕分隔"
        info = self.render_info

        if info is None:
            return

        if isinstance(self.content, SessionBufferControl):
            b = self.content.buffer
            if not b:
                return
            start_lineno = b.start_lineno
            if (start_lineno >= 0) and (start_lineno < b.lineCount - len(info.displayed_lines)):
                b.start_lineno = b.start_lineno + 1
            else:
                b.start_lineno = -1

    def _scroll_up(self) -> None:
        "向上滚屏，处理屏幕分隔"
        info = self.render_info

        if info is None:
            return

        if isinstance(self.content, SessionBufferControl):
            b = self.content.buffer
            if not b:
                return
            start_lineno = b.start_lineno
            if start_lineno > 0:
                b.start_lineno = b.start_lineno - 1
                
            elif start_lineno == 0:
                b.start_lineno = 0

            else:
                b.start_lineno = b.lineCount - len(info.displayed_lines) - 1


class EasternButton(Button):
    "解决增加中文等东亚全宽字符后不对齐问题"

    def _get_text_fragments(self) -> StyleAndTextTuples:
        # 主要改动在这里
        width = self.width - (
            get_cwidth(self.left_symbol) + get_cwidth(self.right_symbol)
        ) - (get_cwidth(self.text) - len(self.text))


        text = (f"{{:^{width}}}").format(self.text)

        def handler(mouse_event: MouseEvent) -> None:
            if (
                self.handler is not None
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.handler()

        return [
            ("class:button.arrow", self.left_symbol, handler),
            #("[SetCursorPosition]", ""),
            ("class:button.text", text, handler),
            ("class:button.arrow", self.right_symbol, handler),
        ]

class EasternMenuContainer(MenuContainer):
    "解决增加中文等东亚全宽字符后不对齐问题"

    def _submenu(self, level: int = 0) -> Window:
        def get_text_fragments() -> StyleAndTextTuples:
            result: StyleAndTextTuples = []
            if level < len(self.selected_menu):
                menu = self._get_menu(level)
                if menu.children:
                    result.append(("class:menu", Border.TOP_LEFT))
                    result.append(("class:menu", Border.HORIZONTAL * (menu.width + 4)))
                    result.append(("class:menu", Border.TOP_RIGHT))
                    result.append(("", "\n"))
                    try:
                        selected_item = self.selected_menu[level + 1]
                    except IndexError:
                        selected_item = -1

                    def one_item(
                        i: int, item: MenuItem
                    ) -> Iterable[OneStyleAndTextTuple]:
                        def mouse_handler(mouse_event: MouseEvent) -> None:
                            if item.disabled:
                                # The arrow keys can't interact with menu items that are disabled.
                                # The mouse shouldn't be able to either.
                                return
                            hover = mouse_event.event_type == MouseEventType.MOUSE_MOVE
                            if (
                                mouse_event.event_type == MouseEventType.MOUSE_UP
                                or hover
                            ):
                                app = get_app()
                                if not hover and item.handler:
                                    app.layout.focus_last()
                                    item.handler()
                                else:
                                    self.selected_menu = self.selected_menu[
                                        : level + 1
                                    ] + [i]

                        if i == selected_item:
                            yield ("[SetCursorPosition]", "")
                            style = "class:menu-bar.selected-item"
                        else:
                            style = ""

                        yield ("class:menu", Border.VERTICAL)
                        if item.text == "-":
                            yield (
                                style + "class:menu-border",
                                f"{Border.HORIZONTAL * (menu.width + 3)}",
                                mouse_handler,
                            )
                        else:
                            # 主要改动在这里，其他地方都未更改.
                            adj_width = menu.width + 3 - (get_cwidth(item.text) - len(item.text))
                            yield (
                                style,
                                f" {item.text}".ljust(adj_width),
                                mouse_handler,
                            )

                        if item.children:
                            yield (style, ">", mouse_handler)
                        else:
                            yield (style, " ", mouse_handler)

                        if i == selected_item:
                            yield ("[SetMenuPosition]", "")
                        yield ("class:menu", Border.VERTICAL)

                        yield ("", "\n")

                    for i, item in enumerate(menu.children):
                        result.extend(one_item(i, item))

                    result.append(("class:menu", Border.BOTTOM_LEFT))
                    result.append(("class:menu", Border.HORIZONTAL * (menu.width + 4)))
                    result.append(("class:menu", Border.BOTTOM_RIGHT))
            return result

        return Window(FormattedTextControl(get_text_fragments), style="class:menu")

@dataclass
class SessionSelectionState:
    start_row: int = -1
    end_row: int = -1
    start_col: int = -1
    end_col: int = -1
    def is_valid(self):
        if self.start_row >= 0 and self.end_row >= 0 and self.start_col >= 0 and self.end_col >= 0:
            if (self.start_row == self.end_row) and (self.start_col == self.end_col):
                return False
            elif self.start_row > self.end_row:
                srow, scol = self.end_row, self.end_col
                erow, ecol = self.start_row, self.start_col
                self.start_row, self.end_row = srow, erow
                self.start_col, self.end_col = scol, ecol
            elif self.start_row == self.end_row and self.start_col > self.end_col:
                scol, ecol = self.end_col, self.start_col
                self.start_col, self.end_col = scol, ecol

            return True

        return False

    @property
    def rows(self):
        if self.is_valid():
            return self.end_row - self.start_row + 1
        else:
            return 0
    

class SessionBuffer:
    def __init__(
        self, 
        name, 
        newline = "\n",
        max_buffered_lines = 10000,
        ) -> None:

        self.name = name
        self.lines = []
        self.newline = newline
        self.isnewline = True
        self.max_buffered_lines = max_buffered_lines
        self.selection = SessionSelectionState(-1, -1, -1, -1)
        self.start_lineno = -1

    def append(self, line: str):
        """
        追加文本到缓冲区。
        当文本以换行符结尾时，会自动添加到缓冲区。
        当文本不以换行符结尾时，会自动添加到上一行。
        """
        newline_after_append = False
        if line.endswith(self.newline):
            line = line.rstrip(self.newline)
            newline_after_append = True
        if not self.newline in line:
            if self.isnewline:
                self.lines.append(line)
            else:
                self.lines[-1] += line

        else:
            lines = line.split(self.newline)
            if self.isnewline:
                self.lines.extend(lines)
            else:
                self.lines[-1] += lines[0]
                self.lines.extend(lines[1:])

        self.isnewline = newline_after_append

        ## limit buffered lines
        if len(self.lines) > self.max_buffered_lines:
            diff = self.max_buffered_lines - len(self.lines)
            del self.lines[:diff]
            ## adjust selection
            if self.selection.start_row >= 0:
                self.selection.start_row -= diff
                self.selection.end_row -= diff

        get_app().invalidate()

    def clear(self):
        self.lines.clear()
        self.selection = SessionSelectionState(-1, -1, -1, -1)

        get_app().invalidate()

    def loadfile(self, filename, encoding = 'utf-8', errors = 'ignore'):
        with open(filename, 'r', encoding = encoding, errors = errors) as fp:
            lines = fp.readlines()
            self.clear()
            self.lines.extend(lines)

            get_app().invalidate()

    @property
    def lineCount(self):
        return len(self.lines)
        
    def getLine(self, lineno):
        if lineno < 0 or lineno >= len(self.lines):
            return ""
        return self.lines[lineno]

    # 获取指定某行到某行的内容。当start未设置时，从首行开始。当end未设置时，到最后一行结束。
    # 注意判断首位顺序逻辑，以及给定参数是否越界
    def getLines(self, start = None, end = None):
        if start is None:
            start = 0
        if end is None:
            end = len(self.lines) - 1
        if start < 0:
            start = 0
        if end >= len(self.lines):
            end = len(self.lines) - 1
        if start > end:
            return []
        return self.lines[start:end+1]

    def selection_range_at_line(self, lineno):
        if self.selection.is_valid():
            if self.selection.rows > 1:
                if lineno == self.selection.start_row:
                    return (self.selection.start_col, len(self.lines[lineno]) - 1)
                elif lineno == self.selection.end_row:
                    return (0, self.selection.end_col)
                elif lineno > self.selection.start_row and lineno < self.selection.end_row:
                    return (0, len(self.lines[lineno]) - 1)

            elif self.selection.rows == 1:
                if lineno == self.selection.start_row:
                    return (self.selection.start_col, self.selection.end_col)

        return None

    def exit_selection(self):
        self.selection = SessionSelectionState(-1, -1, -1, -1)

    def nosplit(self):
        self.start_lineno = -1
        get_app().invalidate()

class SessionBufferControl(UIControl):
    def __init__(self, buffer: Optional[SessionBuffer]) -> None:
        self.buffer = buffer

        # 为MUD显示进行校正的处理，包括对齐校正，换行颜色校正等
        self.FULL_BLOCKS = set("▂▃▅▆▇▄█")
        self.SINGLE_LINES = set("┌└├┬┼┴╭╰─")
        self.DOUBLE_LINES = set("╔╚╠╦╪╩═")
        self.ALL_COLOR_REGX  = re.compile(r"(?:\[[\d;]+m)+")
        self.AVAI_COLOR_REGX = re.compile(r"(?:\[[\d;]+m)+(?!$)")
        self._color_start = ""
        self._color_correction = False
        self._color_line_index = 0

        self._last_click_timestamp = 0

    def reset(self) -> None:
        # Default reset. (Doesn't have to be implemented.)
        pass

    def preferred_width(self, max_available_width: int) -> int | None:
        return None

    def is_focusable(self) -> bool:
        """
        Tell whether this user control is focusable.
        """
        return False


    def width_correction(self, line: str) -> str:
        new_str = []
        for ch in line:
            new_str.append(ch)
            if (east_asian_width(ch) in "FWA") and (wcwidth(ch) == 1):
                if ch in self.FULL_BLOCKS:
                    new_str.append(ch)
                elif ch in self.SINGLE_LINES:
                    new_str.append("─")
                elif ch in self.DOUBLE_LINES:
                    new_str.append("═")
                else:
                    new_str.append(' ')

        return "".join(new_str)
    
    def return_correction(self, line: str):
        return line.replace("\r", "").replace("\x00", "")
    
    def tab_correction(self, line: str):
        return line.replace("\t", " " * Settings.client["tabstop"])

    def line_correction(self, line: str):
        # 处理\r符号（^M）
        line = self.return_correction(line)
        # 处理Tab(\r)符号（^I）
        line = self.tab_correction(line)
        
        # 美化（解决中文英文在Console中不对齐的问题）
        if Settings.client["beautify"]:
            line = self.width_correction(line)

        return line

    def create_content(self, width: int, height: int) -> UIContent:
        """
        Generate the content for this user control.

        Returns a :class:`.UIContent` instance.
        """
        buffer = self.buffer
        if not buffer:
            return UIContent(
                get_line = lambda i: [],
                line_count = 0,
                cursor_position = None
            )

        def get_line(i: int) -> StyleAndTextTuples:
            line = buffer.getLine(i)
            # 颜色校正
            thislinecolors = len(self.AVAI_COLOR_REGX.findall(line))
            if thislinecolors == 0:
                lineno = i - 1
                while lineno >= 0:
                    lastline = buffer.getLine(lineno)
                    allcolors = self.ALL_COLOR_REGX.findall(lastline)
                    
                    if len(allcolors) == 0:
                        lineno = lineno - 1

                    elif len(allcolors) == 1:
                        colors = self.AVAI_COLOR_REGX.findall(lastline)
                        
                        if len(colors) == 1:
                            line = f"{colors[0]}{line}"
                            break

                        else:
                            break

                    else:
                        break

            
            # 其他校正
            line = self.line_correction(line)

            # 处理ANSI标记（生成FormmatedText）
            fragments = to_formatted_text(ANSI(line))

            # 选择内容标识
            selected_fragment = " class:selected "

            # In case of selection, highlight all matches.
            selection_at_line = buffer.selection_range_at_line(i)

            if selection_at_line:
                from_, to = selection_at_line
                # from_ = source_to_display(from_)
                # to = source_to_display(to)

                fragments = explode_text_fragments(fragments)

                if from_ == 0 and to == 0 and len(fragments) == 0:
                    # When this is an empty line, insert a space in order to
                    # visualize the selection.
                    return [(selected_fragment, " ")]
                else:
                    for i in range(from_, to):
                        if i < len(fragments):
                            old_fragment, old_text, *_ = fragments[i]
                            fragments[i] = (old_fragment + selected_fragment, old_text)
                        elif i == len(fragments):
                            fragments.append((selected_fragment, " "))

            return fragments

        content = UIContent(
            get_line = get_line,
            line_count = buffer.lineCount,
            cursor_position = None
        )

        return content

    def mouse_handler(self, mouse_event: MouseEvent):
        """
        Handle mouse events.

        When `NotImplemented` is returned, it means that the given event is not
        handled by the `UIControl` itself. The `Window` or key bindings can
        decide to handle this event as scrolling or changing focus.

        :param mouse_event: `MouseEvent` instance.
        """
        """
        鼠标处理，修改内容包括：
        1. 在CommandLine获得焦点的时候，鼠标对本Control也可以操作
        2. 鼠标双击为选中行
        """
        buffer = self.buffer
        position = mouse_event.position

        # Focus buffer when clicked.
        cur_control = get_app().layout.current_control
        cur_buffer = get_app().layout.current_buffer
        # 这里是修改的内容
        if (cur_control == self) or (cur_buffer and cur_buffer.name == "input"):

            if buffer:
                # Set the selection position.
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    buffer.exit_selection()
                    buffer.selection.start_row = position.y
                    buffer.selection.start_col = position.x

                elif (
                    mouse_event.event_type == MouseEventType.MOUSE_MOVE
                    and mouse_event.button == MouseButton.LEFT
                ):
                    # Click and drag to highlight a selection
                    if buffer.selection.start_row >= 0:
                        buffer.selection.end_row = position.y
                        buffer.selection.end_col = position.x

                elif mouse_event.event_type == MouseEventType.MOUSE_UP:
                    # When the cursor was moved to another place, select the text.
                    # (The >1 is actually a small but acceptable workaround for
                    # selecting text in Vi navigation mode. In navigation mode,
                    # the cursor can never be after the text, so the cursor
                    # will be repositioned automatically.)
                    
                    if buffer.selection.start_row >= 0:
                        buffer.selection.end_row = position.y
                        buffer.selection.end_col = position.x

                        if buffer.selection.start_row == buffer.selection.end_row and buffer.selection.start_col == buffer.selection.end_col:
                            buffer.selection = SessionSelectionState(-1, -1, -1, -1)


                    # Select word around cursor on double click.
                    # Two MOUSE_UP events in a short timespan are considered a double click.
                    double_click = (
                        self._last_click_timestamp
                        and time.time() - self._last_click_timestamp < 0.3
                    )
                    self._last_click_timestamp = time.time()

                    if double_click:
                        buffer.selection.start_row = position.y
                        buffer.selection.start_col = 0
                        buffer.selection.end_row = position.y
                        buffer.selection.end_col = len(buffer.getLine(position.y))

                    get_app().layout.focus("input")

                else:
                    # Don't handle scroll events here.
                    return NotImplemented

        # Not focused, but focusing on click events.
        else:
                return NotImplemented

        return None

class DotDict(dict):
    """
    可以通过点.访问内部key-value对的字典。此类型继承自dict。

    - 由于继承关系，此类型可以使用所有dict可以使用的方法
    - 额外增加的点.访问方法使用示例如下

    示例:
        .. code:: Python

            mydict = DotDict()
            
            # 以下写内容访问等价
            mydict["key1"] = "value1"
            mydict.key1 = "value1"

            # 以下读访问等价
            val = mydict["key1"]
            val = mydict.key1
    """

    def __getattr__(self, __key):
        if (not __key in self.__dict__) and (not __key.startswith("__")):
            return self.__getitem__(__key)

    def __setattr__(self, __name: str, __value):
        if __name in self.__dict__:
            object.__setattr__(self, __name, __value)
        else:
            self.__setitem__(__name, __value)

    def __getstate__(self):
        return self
    
    def __setstate__(self, state):
        self.update(state)


        



import importlib, importlib.util, traceback
from typing import Any
from .settings import Settings
from .extras import DotDict
from .decorators import exception, async_exception, PymudDecorator, print_exception

class PymudMeta(type):
    def __new__(cls, name, bases, attrs):
        decorator_funcs = {}
        for name, value in attrs.items():
            if hasattr(value, "__pymud_decorators__"):
                decorator_funcs[value.__name__] = getattr(value, "__pymud_decorators__", [])
                
        attrs["_decorator_funcs"] = decorator_funcs

        return super().__new__(cls, name, bases, attrs)
    
class ModuleInfo:
    """
    模块管理类。对加载的模块文件进行管理。该类型由Session类进行管理，无需人工创建和干预。

    有关模块的分类和使用的详细信息，请参见 `脚本 <scripts.html>`_

    :param module_name: 模块的名称, 应与 import xxx 语法中的 xxx 保持一致
    :param session: 加载/创建本模块的会话

    """
    def __init__(self, module_name: str, session):
        from .session import Session
        if isinstance(session, Session):
            self.session = session
        self._name = module_name 
        self._ismainmodule = False
        self.load()
       
    def _load(self, reload = False):
        result = True
        if reload:
            self._module = importlib.reload(self._module)
        else:
            self._module = importlib.import_module(self.name)
        self._config = {}
        for attr_name in dir(self._module):
            attr = getattr(self._module, attr_name)
            if isinstance(attr, type) and attr.__module__ == self._module.__name__:
                if (attr_name == "Configuration") or issubclass(attr, IConfig):
                    try:
                        self._config[f"{self.name}.{attr_name}"] = attr(self.session, reload = reload)
                        if not reload:
                            self.session.info(Settings.gettext("configuration_created", self.name, attr_name))
                        else:
                            self.session.info(Settings.gettext("configuration_recreated", self.name, attr_name))

                    except Exception as e:
                        result = False
                        self.session.error(Settings.gettext("configuration_fail", self.name, attr_name, e))
        self._ismainmodule = (self._config != {})
        return result
    
    def _unload(self):
        from .objects import BaseObject, Command
        for key, config in self._config.items():
            if isinstance(config, Command):
                # Command 对象在从会话中移除时，自动调用其 unload 系列方法，因此不能产生递归
                self.session.delObject(config)
            
            else:

                if hasattr(config, "__unload__"):
                    unload = getattr(config, "__unload__", None)
                    if callable(unload): unload()
                    
                if hasattr(config, "unload"):
                    unload = getattr(config, "unload", None)
                    if callable(unload): unload()

                if isinstance(config, BaseObject):
                    self.session.delObject(config)

            del config
        self._config.clear()

    def load(self):
        "加载模块内容"
        if self._load():
            self.session.info(f"{Settings.gettext('entity_module' if self.ismainmodule else 'non_entity_module')} {self.name} {Settings.gettext('load_ok')}")
        else:
            self.session.info(f"{Settings.gettext('entity_module' if self.ismainmodule else 'non_entity_module')} {self.name} {Settings.gettext('load_fail')}")

    def unload(self):
        "卸载模块内容"
        self._unload()
        self._loaded = False
        self.session.info(f"{Settings.gettext('entity_module' if self.ismainmodule else 'non_entity_module')} {self.name} {Settings.gettext('unload_ok')}")

    def reload(self):
        "模块文件更新后调用，重新加载已加载的模块内容"
        self._unload()
        self._load(reload = True)
        self.session.info(f"{Settings.gettext('entity_module' if self.ismainmodule else 'non_entity_module')} {self.name} {Settings.gettext('reload_ok')}")

    @property
    def name(self):
        "只读属性，模块名称"
        return self._name

    @property
    def module(self):
        "只读属性，模块文件的 ModuleType 对象"
        return self._module
    
    @property
    def config(self):
        "只读字典属性，根据模块文件 ModuleType 对象创建的其中名为 Configuration 的类型或继承自 IConfig 的子类型实例（若有）"
        return self._config
    
    @property
    def ismainmodule(self):
        "只读属性，区分是否主模块（即包含具体config的模块）"
        return self._ismainmodule

class IConfigBase(metaclass = PymudMeta):
    """
    用于支持对装饰器写法对象进行管理的基础类。
    该类型相当于原来的IConfig类，唯一区别时，模块加载时，不会对本类型创建实例对象。
    主要用于对插件中定义的Command提供装饰器写法支持，因为这些Command是在会话构建时创建，因此不能在模块加载时自动创建，也就不能继承自IConfig。
    """
    def __init__(self, session, *args, **kwargs):
        from .session import Session
        from .objects import Alias, Trigger, Timer, GMCPTrigger
        if isinstance(session, Session):
            self.session = session
        self.__inline_objects__ = DotDict()

        if hasattr(self, "_decorator_funcs"):
            deco_funcs = getattr(self, "_decorator_funcs")
            for func_name, decorators in deco_funcs.items():
                func = getattr(self, func_name)
                for deco in decorators:
                    if isinstance(deco, PymudDecorator):
                        if deco.type == "alias":
                            #patterns = deco.kwargs.pop("patterns")
                            ali = Alias(self.session, *deco.args, **deco.kwargs, onSuccess = func)
                            self.__inline_objects__[ali.id] = ali
                        
                        elif deco.type == "trigger":
                            #patterns = deco.kwargs.pop("patterns")
                            tri = Trigger(self.session, *deco.args, **deco.kwargs, onSuccess = func)
                            self.__inline_objects__[tri.id] = tri

                        elif deco.type == "timer":
                            tim = Timer(self.session, *deco.args, **deco.kwargs, onSuccess = func)
                            self.__inline_objects__[tim.id] = tim

                        elif deco.type == "gmcp":
                            gmcp = GMCPTrigger(self.session, name = deco.kwargs.get("id"), *deco.args, **deco.kwargs, onSuccess = func)
                            self.__inline_objects__[gmcp.id] = gmcp

    def __unload__(self):
        from .objects import BaseObject
        if self.session:
            self.session.delObjects(self.__inline_objects__)
            if isinstance(self, BaseObject):
                self.session.delObject(self)

    @property
    def objs(self) -> DotDict:
        "返回内联自动创建的对象字典"
        return self.__inline_objects__
    
    def obj(self, obj_id: str):
        "根据对象ID返回内联自动创建的对象"
        return self.__inline_objects__.get(obj_id, None) # type: ignore

class IConfig(IConfigBase):
    """
    用于提示PyMUD应用是否自动创建该配置类型的基础类。
    
    继承 IConfig 类型让应用自动管理该类型，唯一需要的是，构造函数中，仅存在一个必须指定的参数 Session。

    在应用自动创建 IConfig 实例时，除 session 参数外，还会传递一个 reload 参数 （bool类型），表示是首次加载还是重新加载特性。
    可以从kwargs 中获取该参数，并针对性的设计相应代码。例如，重新加载相关联的其他模块等。
    """
        
class Plugin:
    """
    插件管理类。对加载的插件文件进行管理。该类型由PyMudApp进行管理，无需人工创建。

    有关插件的详细信息，请参见 `插件 <plugins.html>`_

    :param name: 插件的文件名, 如'myplugin.py'
    :param location: 插件所在的目录。自动加载的插件包括PyMUD包目录下的plugins目录以及当前目录下的plugins目录

    """
    def __init__(self, name, location):
        self._plugin_file = name
        self._plugin_loc  = location

        self.reload()

    def reload(self):
        "加载/重新加载插件对象"
        #del self.modspec, self.mod
        self.modspec = importlib.util.spec_from_file_location(self._plugin_file[:-3], self._plugin_loc)
        if self.modspec and self.modspec.loader:
            self.mod     = importlib.util.module_from_spec(self.modspec)
            self.modspec.loader.exec_module(self.mod)
        
            self._app_init = self._load_mod_function("PLUGIN_PYMUD_START")
            self._session_create = self._load_mod_function("PLUGIN_SESSION_CREATE")
            self._session_destroy = self._load_mod_function("PLUGIN_SESSION_DESTROY")
            self._app_destroy = self._load_mod_function("PLUGIN_PYMUD_DESTROY")

        else:
            raise FileNotFoundError(Settings.gettext("exception_plugin_file_not_found", self._plugin_file))
        
    def _load_mod_function(self, func_name):
        # 定义一个默认函数，当插件文件中未定义指定名称的函数时，返回该函数
        # 该函数接受任意数量的位置参数和关键字参数，但不执行任何操作
        def default_func(*args, **kwargs):
            pass

        result = default_func
        if func_name in self.mod.__dict__:
            func = self.mod.__dict__[func_name]
            if callable(func):
                result = func
        return result

    @property
    def name(self):
        "插件名称，由插件文件中的 PLUGIN_NAME 常量定义"
        return self.mod.__dict__["PLUGIN_NAME"]
    
    @property
    def desc(self):
        "插件描述，由插件文件中的 PLUGIN_DESC 常量定义"
        return self.mod.__dict__["PLUGIN_DESC"]
    
    @property
    def help(self):
        "插件帮助，由插件文件中的文档字符串定义"
        return self.mod.__doc__
    
    def onAppInit(self, app):
        """
        PyMUD应用启动时对插件执行的操作，由插件文件中的 PLUGIN_PYMUD_START 函数定义

        :param app: 启动的 PyMudApp 对象实例
        """
        self._app_init(app)

    def onSessionCreate(self, session):
        """
        新会话创建时对插件执行的操作，由插件文件中的 PLUGIN_SESSION_CREATE 函数定义

        :param session: 新创建的会话对象实例
        """
        try:
            self._session_create(session)
        except Exception as e:
            print_exception(session, e)
            

    def onSessionDestroy(self, session):
        """
        会话关闭时（注意不是断开）对插件执行的操作，由插件文件中的 PLUGIN_SESSION_DESTROY 函数定义

        :param session: 所关闭的会话对象实例
        """
        try:
            self._session_destroy(session)
        except Exception as e:
            print_exception(session, e)

    def onAppDestroy(self, app):
        """
        PyMUD应用关闭时对插件执行的操作，由插件文件中的 PLUGIN_PYMUD_DESTROY 函数定义
        :param app: 关闭的 PyMudApp 对象实例
        """
        self._app_destroy(app)

    def __getattr__(self, __name: str) -> Any:
        if hasattr(self.mod, __name):
            return self.mod.__getattribute__(__name)


import os, sys, json, platform, shutil, logging, argparse, locale
from pathlib import Path
from .pymud import PyMudApp
from .settings import Settings

CFG_TEMPLATE = {
    "language" : "chs",                             # 语言设置，默认为简体中文

    "client": {
        "buffer_lines"      : 5000,                 # 保留缓冲行数

        "interval"          : 10,                   # 在自动执行中，两次命令输入中的间隔时间（ms）
        "auto_connect"      : True,                 # 创建会话后，是否自动连接
        "auto_reconnect"    : False,                # 在会话异常断开之后，是否自动重连
        "reconnect_wait"    : 15,                   # 自动重连等待的时间（秒数）
        "var_autosave"      : True,                 # 断开时自动保存会话变量
        "var_autoload"      : True,                 # 初始化时自动加载会话变量

        "beautify"          : True,                 # 专门为解决控制台下PKUXKX字符画对不齐的问题
        "history_records"   : 500, 

        "status_divider"    : False,                 # 是否显示状态栏的分隔线
        "status_display"    : 1,                    # 状态窗口显示情况设置，0-不显示，1-显示在下方，2-显示在右侧
        "status_height"     : 4,                    # 下侧状态栏的高度
        "status_width"      : 30,                   # 右侧状态栏的宽度
        

    },

    "sessions" : {
        "pkuxkx" : {
            "host" : "mud.pkuxkx.net",
            "port" : "8081",
            "encoding" : "utf8",
            "autologin" : "{0};{1}",
            "default_script": "examples",
            "chars" : {
                "display_title" : ["yourid", "yourpassword", ""],
            }
        }
    },
    "keys" : {
        "f3"    : "#ig",
        "f4"    : "#clear",
        "f11"   : "#close",
        "f12"   : "#exit",
    }
}

def detect_system_language():
    """
    检测系统语言，返回中文或英文"
    """
    lang = "chs"
    try:
        value = locale.getlocale()[0]
        if value and (value.lower().startswith("zh") or value.lower().startswith("chinese")):  # 中文
            lang = "chs"
        else:
            lang = "eng"
    except Exception as e:
        # default is chs
        pass

    return lang

def init_pymud_env(args):
    system = "unknown"
    lang = detect_system_language()
    if lang == "chs":
        print(f"欢迎使用PyMUD, 版本{Settings.__version__}. 使用PyMUD时, 建议建立一个新目录（任意位置），并将自己的脚本以及配置文件放到该目录下.")
        print("即将开始为首次运行初始化环境...")
        
        dir = args.dir
        if dir:
            if dir == ".":
                dir_msg = "当前目录"
            else:
                dir_msg = f": {dir}"
            print(f"你已经指定了创建脚本的目录为{dir_msg}")
            dir = Path(dir)
        else:
            dir = Path.home().joinpath('pkuxkx')

            system = platform.system().lower()
            dir_enter = input(f"检测到当前系统为 {system}, 请指定游戏脚本的目录（若目录不存在会自动创建），直接回车表示使用默认值 [{dir}]:")
            if dir_enter:
                dir = Path(dir_enter)
                        
        if dir.exists() and dir.is_dir():
            print(f'检测到给定目录 {dir} 已存在，切换至此目录...')
        else:
            print(f'检测到给定目录 {dir} 不存在，正在创建并切换至目录...')
            dir.mkdir()
        
        os.chdir(dir)

        if os.path.exists('pymud.cfg'):
            print(f'检测到脚本目录下已存在pymud.cfg文件，将直接使用此文件进入PyMUD...')
        else:
            print(f'检测到脚本目录下不存在pymud.cfg文件，将使用默认内容创建该配置文件...')
            with open('pymud.cfg', mode = 'x') as fp:
                fp.writelines(json.dumps(CFG_TEMPLATE, indent = 4))

        if not os.path.exists('examples.py'):
            from pymud import pkuxkx
            module_dir = pkuxkx.__file__
            shutil.copyfile(module_dir, 'examples.py')
            print(f'已将样例脚本拷贝至脚本目录，并加入默认配置文件')

        print(f"后续可自行修改 {dir} 目录下的 pymud.cfg 文件以进行配置。")
        if system == "windows":
            print(f"后续运行PyMUD， 请在 {dir} 目录下键入命令： python -m pymud，或直接使用快捷命令 pymud")
        else:
            print(f"后续运行PyMUD， 请在 {dir} 目录下键入命令： python3 -m pymud，或直接使用快捷命令 pymud")

        input('所有内容已初始化完毕, 请按回车进入PyMUD.')

    else:
        print(f"Welcome to PyMUD, version {Settings.__version__}. When using pymud, it is suggested that a new folder should be created (in any place), and the cfg configuration and all the scripts have been placed in the directory.")
        print("Starting to initialize the environment for the first time...")
        dir = args.dir
        if dir:
            if dir == ".":
                dir_msg = "current directory"
            else:
                dir_msg = f": {dir}"
            print(f"You have specified the directory to create the script as {dir_msg}")
            dir = Path(dir)
        else:
            dir = Path.home().joinpath('pkuxkx')

            system = platform.system().lower()
            dir_enter = input(f"Detected the current system is {system}, please specify the directory of the game script (if the directory does not exist, it will be automatically created), press Enter to use the default value [{dir}]:")
            if dir_enter:
                dir = Path(dir_enter)
                        
        if dir.exists() and dir.is_dir():
            print(f'Detected that the given directory {dir} already exists, switching to this directory...')
        else:
            print(f'Detected that the given directory {dir} does not exist, creating and switching to the directory...')
            dir.mkdir()
        
        os.chdir(dir)

        if os.path.exists('pymud.cfg'):
            print(f'Detected that the pymud.cfg file already exists in the script directory, entering PyMUD directly using this file...')
        else:
            print(f'Detected that the pymud.cfg file does not exist in the script directory, creating the configuration file using the default content...')
            with open('pymud.cfg', mode = 'x') as fp:
                CFG_TEMPLATE["language"] = "eng"
                fp.writelines(json.dumps(CFG_TEMPLATE, indent = 4))

        if not os.path.exists('examples.py'):
            from pymud import pkuxkx
            module_dir = pkuxkx.__file__
            shutil.copyfile(module_dir, 'examples.py')
            print(f'The sample script has been copied to the script directory and added to the default configuration file')

        print(f"Afterwards, you can modify the pymud.cfg file in the {dir} directory for configuration.")
        if system == "windows":
            print(f"Afterwards, please type the command 'python -m pymud' in the {dir} directory to run PyMUD, or use the shortcut command pymud")
        else:
            print(f"Afterwards, please type the command 'python3 -m pymud' in the {dir} directory to run PyMUD, or use the shortcut command pymud")

        input('Press Enter to enter PyMUD.')

    startApp(args)

def startApp(args):
    startup_path = Path(args.startup_dir).resolve()
    sys.path.append(f"{startup_path}")
    os.chdir(startup_path)

    if args.debug:
        logging.basicConfig(level = logging.NOTSET,
            format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            datefmt = '%m-%d %H:%M',
            filename = args.logfile,
            filemode = 'a' if args.filemode else 'w',
            encoding = "utf-8"
            )
        
    else:
        logging.basicConfig(level = logging.NOTSET,
            format = '%(asctime)s %(name)-12s: %(message)s',
            datefmt = '%m-%d %H:%M',
            handlers = [logging.NullHandler()],
            )

    cfg = startup_path.joinpath("pymud.cfg")
    cfg_data = None
    if os.path.exists(cfg):
        with open(cfg, "r", encoding="utf8", errors="ignore") as fp:
            cfg_data = json.load(fp)

    app = PyMudApp(cfg_data)
    app.run()

def main():
    parser = argparse.ArgumentParser(prog = "pymud", description = "PyMUD命令行参数帮助")
    subparsers = parser.add_subparsers(help = 'init用于初始化运行环境')

    par_init = subparsers.add_parser('init', description = '初始化pymud运行环境, 包括建立脚本目录, 创建默认配置文件, 创建样例脚本等.')
    par_init.add_argument('-d', '--dir', dest = 'dir', metavar = 'dir', type = str, default = '.', help = '指定构建脚本目录的名称, 不指定时会根据操作系统选择不同默认值')
    par_init.set_defaults(func = init_pymud_env)

    parser.add_argument('-d', '--debug', dest = 'debug', action = 'store_true', default = False, help = '指定以调试模式进入PyMUD。此时，系统log等级将设置为logging.NOTSET, 所有log数据均会被记录。默认不启用。')
    parser.add_argument('-l', '--logfile', dest = 'logfile', metavar = 'logfile', default = 'pymud.log', help = '指定调试模式下记录文件名，不指定时，默认为当前目录下的pymud.log')
    parser.add_argument('-a', '--appendmode', dest = 'filemode', action = 'store_true', default = True, help = '指定log文件的访问模式是否为append尾部添加模式，默认为True。当为False时，使用w模式，即每次运行清空之前记录')
    parser.add_argument('-s', '--startup_dir', dest = 'startup_dir', metavar = 'startup_dir', default = '.', help = '指定启动目录，默认为当前目录。使用该参数可以在任何目录下，通过指定脚本目录来启动')

    args=parser.parse_args()

    if hasattr(args, 'func'):
        args.func(args)
    else:
        startApp(args)

if __name__ == "__main__":
    main()


from .settings import Settings
from .pymud import PyMudApp
from .modules import IConfigBase, IConfig, PymudMeta
from .objects import CodeBlock, Alias, SimpleAlias, Trigger, SimpleTrigger, Command, SimpleCommand, Timer, SimpleTimer, GMCPTrigger
from .extras import DotDict
from .session import Session
from .logger import Logger
from .main import main
from .decorators import exception, async_exception, PymudDecorator, alias, trigger, timer, gmcp

__all__ = [
    "PymudMeta", "IConfigBase", "IConfig", "PyMudApp", "Settings", "CodeBlock", 
    "Alias", "SimpleAlias", "Trigger", "SimpleTrigger", 
    "Command", "SimpleCommand", "Timer", "SimpleTimer", 
    "GMCPTrigger", "Session", "DotDict", "Logger", "main",
    "exception", "async_exception", "PymudDecorator", "alias", "trigger", "timer", "gmcp",
]


import asyncio, functools, re, os, webbrowser, threading
from datetime import datetime
from pathlib import Path
from prompt_toolkit.shortcuts import set_title, radiolist_dialog
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit import HTML
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.application import Application
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import ConditionalContainer, Float, VSplit, HSplit, Window, WindowAlign, ScrollbarMargin, NumberedMargin
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import Label, TextArea
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.cursor_shapes import CursorShape
from prompt_toolkit.key_binding import KeyPress, KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.filters import (
    Condition,
    is_true,
    to_filter,
)
from prompt_toolkit.formatted_text import (
    Template,
)
from prompt_toolkit.layout.processors import (
    DisplayMultipleCursors,
    HighlightSearchProcessor,
    HighlightSelectionProcessor,
)
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from wcwidth import wcwidth, wcswidth

from .objects import CodeBlock
from .extras import SessionBuffer, SessionBufferControl, EasternMenuContainer, VSplitWindow, DotDict, MenuItem
from .modules import Plugin
from .session import Session
from .settings import Settings
from .dialogs import MessageDialog, WelcomeDialog, QueryDialog, NewSessionDialog, LogSelectionDialog

from enum import Enum

class STATUS_DISPLAY(Enum):
    NONE = 0
    HORIZON = 1
    VERTICAL = 2
    FLOAT = 3

class PyMudApp:
    """
    PYMUD程序管理主对象，对窗体、操作及所有会话进行管理。

    PyMudApp对象不需要手动创建，在命令行中执行 ``python -m pymud`` 时会自动创建对象实例。

    参数： 
        - ``cfg_data``: 替代配置数据，由本地pymud.cfg文件读取，用于覆盖settings.py中的默认Settings数据

    可替代字典: 含义请查阅 `应用配置及本地化 <settings.html>`_
        - sessions: 用于创建菜单栏会话的字典
        - client: 用于配置客户端属性的字典
        - text: 用于各默认显示文字内容的字典
        - server: 用于服务器选项的配置字典
        - styles: 用于显示样式的定义字典
        - keys: 用于快捷键定义的字典

    *替代配置按不同的dict使用dict.update进行更新覆盖，因此可以仅指定需替代的部分。*
    """

    def __init__(self, cfg_data = None) -> None:
        """
        构造PyMudApp对象实例，并加载替代配置。
        """

        from .i18n import i18n_LoadLanguage, i18n_ListAvailableLanguages
        # 加载默认chs语言内容，以防翻译不完整时，默认使用中文替代
        i18n_LoadLanguage("chs")

        if cfg_data and isinstance(cfg_data, dict):
            # load language from 
            language = Settings.language
            if "language" in cfg_data.keys():
                if cfg_data["language"] in i18n_ListAvailableLanguages() and cfg_data["language"] != "chs":
                    language = cfg_data["language"]
                    i18n_LoadLanguage(language)
            
            for key in cfg_data.keys():
                if key == "sessions":
                    Settings.sessions = cfg_data[key]
                elif key == "client":
                    Settings.client.update(cfg_data[key])
                elif key == "text":
                    Settings.text.update(cfg_data[key])
                elif key == "server":
                    Settings.server.update(cfg_data[key])
                elif key == "styles":
                    Settings.styles.update(cfg_data[key])
                elif key == "keys":
                    Settings.keys.update(cfg_data[key])
                elif key == "language":
                    Settings.language = cfg_data[key]

        self._mouse_support = True
        self._plugins  = DotDict()              # 增加 插件 字典
        self._globals  = DotDict()              # 增加所有session使用的全局变量
        self._onTimerCallbacks = dict()
        self.sessions = {}
        self.current_session = None
        self.status_display = STATUS_DISPLAY(Settings.client["status_display"])

        self.keybindings = KeyBindings()
        self.keybindings.add(Keys.PageUp, is_global = True)(self.page_up)
        self.keybindings.add(Keys.PageDown, is_global = True)(self.page_down)
        self.keybindings.add(Keys.ControlZ, is_global = True)(self.hide_history)
        self.keybindings.add(Keys.ControlC, is_global = True)(self.copy_selection)      # Control-C 复制文本
        self.keybindings.add(Keys.ControlR, is_global = True)(self.copy_selection)      # Control-R 复制带有ANSI标记的文本（适用于整行复制）
        self.keybindings.add(Keys.Right, is_global = True)(self.complete_autosuggest)   # 右箭头补完建议
        self.keybindings.add(Keys.Backspace)(self.delete_selection)
        self.keybindings.add(Keys.ControlLeft, is_global = True)(self.change_session)   # Control-左右箭头切换当前会话
        self.keybindings.add(Keys.ControlRight, is_global = True)(self.change_session)
        self.keybindings.add(Keys.ShiftLeft, is_global = True)(self.change_session)    # Shift-左右箭头切换当前会话
        self.keybindings.add(Keys.ShiftRight, is_global = True)(self.change_session)   # 适配 MacOS系统
        self.keybindings.add(Keys.F1, is_global=True)(lambda event: webbrowser.open(Settings.__website__))
        self.keybindings.add(Keys.F2, is_global=True)(self.toggle_mousesupport)

        used_keys = [Keys.PageUp, Keys.PageDown, Keys.ControlZ, Keys.ControlC, Keys.ControlR, Keys.Up, Keys.Down, Keys.Left, Keys.Right, Keys.ControlLeft, Keys.ControlRight, Keys.Backspace, Keys.Delete, Keys.F1, Keys.F2]

        for key, binding in Settings.keys.items():
            if (key not in used_keys) and binding and isinstance(binding, str):
                self.keybindings.add(key, is_global = True)(self.custom_key_press)

        self.initUI()

        # 对剪贴板进行处理，经测试，android下的termux中，pyperclip无法使用，因此要使用默认的InMemoryClipboard
        clipboard = None
        try:
            clipboard = PyperclipClipboard()
            clipboard.set_text("test pyperclip")
            clipboard.set_text("")
        except:
            clipboard = None

        self.app = Application(
            layout = Layout(self.root_container, focused_element=self.commandLine),
            enable_page_navigation_bindings=True,
            style=self.style,
            mouse_support=to_filter(self._mouse_support),
            full_screen=True,
            color_depth=ColorDepth.TRUE_COLOR,
            clipboard=clipboard,
            key_bindings=self.keybindings,
            cursor=CursorShape.BLINKING_UNDERLINE
        )

        set_title("{} {}".format(Settings.__appname__, Settings.__version__))
        self.set_status(Settings.text["welcome"])

        self.loggers = dict()           # 所有记录字典
        self.showLog = False            # 是否显示记录页
        self.logFileShown = ''          # 记录页显示的记录文件名
        #self.logSessionBuffer = SessionBuffer()
        #self.logSessionBuffer.name = "LOGBUFFER"
        self.logSessionBuffer = SessionBuffer("LOGBUFFER", max_buffered_lines = -1)

        self.load_plugins()

    async def onSystemTimerTick(self):
        while True:
            await asyncio.sleep(1)
            self.app.invalidate()
            for callback in self._onTimerCallbacks.values():
                if callable(callback):
                    callback()

    def addTimerTickCallback(self, name, func):
        '注册一个系统定时器回调，每1s触发一次。指定name为回调函数关键字，func为回调函数。'
        if callable(func) and (not name in self._onTimerCallbacks.keys()):
            self._onTimerCallbacks[name] = func

    def removeTimerTickCallback(self, name):
        '从系统定时器回调中移除一个回调函数。指定name为回调函数关键字。'
        if name in self._onTimerCallbacks.keys():
            self._onTimerCallbacks.pop(name)

    def initUI(self):
        """初始化UI界面"""
        self.style = Style.from_dict(Settings.styles)
        self.status_message = ""
        self.showHistory = False
        self.wrap_lines  = True

        self.commandLine = TextArea(
            prompt=self.get_input_prompt, 
            multiline = False, 
            accept_handler = self.enter_pressed, 
            height=D(min=1), 
            auto_suggest = AutoSuggestFromHistory(), 
            focus_on_click=True,
            name = "input",
            )

        self.status_bar = VSplit(
            [
                Window(FormattedTextControl(self.get_statusbar_text), style="class:status", align = WindowAlign.LEFT),
                Window(FormattedTextControl(self.get_statusbar_right_text), style="class:status.right", width = D(preferred=40), align = WindowAlign.RIGHT),
            ],
            height = 1,
            style ="class:status"
            )

        # 增加状态窗口显示
        self.statusView = FormattedTextControl(
            text = self.get_statuswindow_text,
            show_cursor=False
        )

        self.consoleView = SessionBufferControl(
            buffer = None,
            )
        

        self.console = VSplitWindow(
            content = self.consoleView,
            width = D(preferred = Settings.client["naws_width"]),
            height = D(preferred = Settings.client["naws_height"]),
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            #left_margins=[NumberedMargin()],
            #right_margins=[ScrollbarMargin(True)],   
            style="class:text-area"
            )

        console_with_bottom_status = ConditionalContainer(
            content = HSplit(
                [
                    self.console,
                    ConditionalContainer(content = Window(char = "—", height = 1), filter = Settings.client["status_divider"]),
                    #Window(char = "—", height = 1),
                    Window(content = self.statusView, height = Settings.client["status_height"]),
                ]
            ),
            filter = to_filter(self.status_display == STATUS_DISPLAY.HORIZON)
        )


        console_with_right_status = ConditionalContainer(
            content = VSplit(
                [
                    self.console,
                    ConditionalContainer(content = Window(char = "|", width = 1), filter = Settings.client["status_divider"]),
                    Window(content = self.statusView, width = Settings.client["status_width"]),
                ]
            ),
            filter = to_filter(self.status_display == STATUS_DISPLAY.VERTICAL)
        )

        console_without_status = ConditionalContainer(
            content = self.console,
            filter = to_filter(self.status_display == STATUS_DISPLAY.NONE)
        )

        body = HSplit(
            [
                console_without_status,
                console_with_right_status,
                console_with_bottom_status
            ]
        )

        fill = functools.partial(Window, style="class:frame.border")
        top_row_with_title = VSplit(
            [
                #fill(width=1, height=1, char=Border.TOP_LEFT),
                fill(char = "\u2500"),
                fill(width=1, height=1, char="|"),
                # Notice: we use `Template` here, because `self.title` can be an
                # `HTML` object for instance.
                Label(
                    lambda: Template(" {} ").format(self.get_frame_title),
                    style="class:frame.label",
                    dont_extend_width=True,
                ),
                fill(width=1, height=1, char="|"),
                fill(char = "\u2500"),
                #fill(width=1, height=1, char=Border.TOP_RIGHT),
            ],
            height=1,
        )

        new_body = HSplit([
            top_row_with_title,
            body,
            fill(height = 1, char = "\u2500"),
        ])

        #self.console_frame = Frame(body = body, title = self.get_frame_title)

        self.body = HSplit([
                new_body,
                #self.console_frame,
                self.commandLine,
                self.status_bar
            ])

        self.root_container = EasternMenuContainer(
            body = self.body,
            menu_items=[
                MenuItem(
                    Settings.gettext("world"),
                    children=self.create_world_menus(),
                ),
                MenuItem(
                    Settings.gettext("session"),
                    children=[
                        MenuItem(Settings.gettext("disconnect"), handler = self.act_discon),
                        MenuItem(Settings.gettext("connect"), handler = self.act_connect),
                        MenuItem(Settings.gettext("closesession"), handler = self.act_close_session),
                        MenuItem(Settings.gettext("autoreconnect"), handler = self.act_autoreconnect),
                        MenuItem("-", disabled=True),
                        MenuItem(Settings.gettext("nosplit"), handler = self.act_nosplit),
                        MenuItem(Settings.gettext("echoinput"), handler = self.act_echoinput),
                        MenuItem(Settings.gettext("beautify"), handler = self.act_beautify),
                        MenuItem(Settings.gettext("copy"), handler = self.act_copy),
                        MenuItem(Settings.gettext("copyraw"), handler = self.act_copyraw),
                        MenuItem(Settings.gettext("clearsession"), handler = self.act_clearsession),
                        MenuItem("-", disabled=True),
                        
                        MenuItem(Settings.gettext("reloadconfig"), handler = self.act_reload),
                    ]
                ),

                # MenuItem(
                #     Settings.text["layout"],
                #     children = [
                #         MenuItem(Settings.text["hide"], handler = functools.partial(self.act_change_layout, False)),
                #         MenuItem(Settings.text["horizon"], handler = functools.partial(self.act_change_layout, True)),
                #         MenuItem(Settings.text["vertical"], handler = functools.partial(self.act_change_layout, True)),
                #     ]
                # ),

                MenuItem(
                    Settings.gettext("help"),
                    children=[
                        MenuItem(Settings.gettext("about"), handler = self.act_about)
                    ]
                ),

                MenuItem(
                    "",    # 增加一个空名称MenuItem，单机后焦点移动至命令行输入处，阻止右侧空白栏点击响应
                    handler = lambda : self.app.layout.focus(self.commandLine)
                )
            ],
            floats=[
                Float(
                    xcursor=True,
                    ycursor=True,
                    content=CompletionsMenu(max_height=16, scroll_offset=1)
                )
            ],
        )

    def create_world_menus(self):
        "创建世界子菜单，其中根据本地pymud.cfg中的有关配置创建会话有关子菜单"
        menus = []
        menus.append(MenuItem(f'{Settings.gettext("new_session")}...', handler = self.act_new))
        menus.append(MenuItem("-", disabled=True))

        ss = Settings.sessions

        for key, site in ss.items():
            menu = MenuItem(key)
            for name in site["chars"].keys():
                sub = MenuItem(name, handler = functools.partial(self._quickHandleSession, key, name)) # type: ignore
                menu.children.append(sub)
            menus.append(menu)

        menus.append(MenuItem("-", disabled=True))
        menus.append(MenuItem(Settings.gettext("show_log"), handler = self.show_logSelectDialog))
        menus.append(MenuItem("-", disabled=True))
        menus.append(MenuItem(Settings.gettext("exit"), handler=self.act_exit))

        return menus

    def invalidate(self):
        "刷新显示界面"
        self.app.invalidate()

    def scroll(self, lines = 1):
        "内容滚动指定行数，小于0为向上滚动，大于0为向下滚动"
        if self.current_session:
            s = self.current_session
            b = s.buffer
        elif self.showLog:
            b = self.logSessionBuffer
        else:
            b = None
            
        if isinstance(b, Buffer):
            if lines < 0:
                b.cursor_up(-1 * lines)
            elif lines > 0:
                b.cursor_down(lines)

    def page_up(self, event: KeyPressEvent) -> None:
        "快捷键PageUp: 用于向上翻页。翻页页数为显示窗口行数的一半减去一行。"
        #lines = (self.app.output.get_size().rows - 5) // 2 - 1
        lines = self.get_height() // 2 - 1
        self.scroll(-1 * lines)

    def page_down(self, event: KeyPressEvent) -> None:
        "快捷键PageDown: 用于向下翻页。翻页页数为显示窗口行数的一半减去一行。"
        #lines = (self.app.output.get_size().rows - 5) // 2 - 1
        lines = self.get_height() // 2 - 1
        self.scroll(lines)

    def custom_key_press(self, event: KeyPressEvent):
        "自定义快捷键功能实现，根据keys字典配置在当前会话执行指定指令"
        if (len(event.key_sequence) == 1) and (event.key_sequence[-1].key in Settings.keys.keys()):
            cmd = Settings.keys[event.key_sequence[-1].key]
            if self.current_session:
                self.current_session.exec_command(cmd)

    def hide_history(self, event: KeyPressEvent) -> None:
        """快捷键Ctrl+Z: 关闭历史行显示"""
        self.act_nosplit()

    def copy_selection(self, event: KeyPressEvent)-> None:
        """快捷键Ctrl+C/Ctrl+R: 复制选择内容。根据按键不同选择文本复制方式和RAW复制方式"""
        if event.key_sequence[-1].key == Keys.ControlC:
            self.copy()
        elif event.key_sequence[-1].key == Keys.ControlR:
            self.copy(raw = True)

    def delete_selection(self, event: KeyPressEvent):
        event.key_sequence
        b = event.current_buffer
        if b.selection_state:
            event.key_processor.feed(KeyPress(Keys.Delete), first=True)
        else:
            b.delete_before_cursor(1)

    def complete_autosuggest(self, event: KeyPressEvent):
        """快捷键右箭头→: 自动完成建议"""
        b = event.current_buffer
        if (b.cursor_position == len(b.text)) and b.auto_suggest:
            s = b.auto_suggest.get_suggestion(b, b.document)
            if s:
                b.insert_text(s.text, fire_event=False)
        else:
            b.cursor_right()

    def change_session(self, event: KeyPressEvent):
        """快捷键Ctrl/Shift+左右箭头: 切换会话"""
        if self.current_session:
            current = self.current_session.name
            keys = list(self.sessions.keys())
            idx = keys.index(current)
            count = len(keys)

            if (event.key_sequence[-1].key == Keys.ControlRight) or (event.key_sequence[-1].key == Keys.ShiftRight):
                if idx < count - 1:
                    new_key = keys[idx+1]
                    self.activate_session(new_key)

                elif (idx == count -1) and self.showLog:
                    self.showLogInTab()

            elif (event.key_sequence[-1].key == Keys.ControlLeft) or (event.key_sequence[-1].key == Keys.ShiftLeft):
                if idx > 0:
                    new_key = keys[idx-1]
                    self.activate_session(new_key)

        else:
            if self.showLog:
                if (event.key_sequence[-1].key == Keys.ControlRight) or (event.key_sequence[-1].key == Keys.ShiftRight):
                    keys = list(self.sessions.keys())
                    if len(keys) > 0:
                        new_key = keys[-1]
                        self.activate_session(new_key)

    def toggle_mousesupport(self, event: KeyPressEvent):
        """快捷键F2: 切换鼠标支持状态。用于远程连接时本地复制命令执行操作"""
        self._mouse_support = not self._mouse_support
        if self._mouse_support:
            self.app.renderer.output.enable_mouse_support()
        else:
            self.app.renderer.output.disable_mouse_support()

    def copy(self, raw = False):
        """
        复制会话中的选中内容

        :param raw: 指定采取文本模式还是ANSI格式模式

        ``注意: 复制的内容仅存在于运行环境的剪贴板中。若使用ssh远程，该复制命令不能访问本地剪贴板。``
        """
        
        b = self.consoleView.buffer
        if b and b.selection.is_valid():
            if not raw:
                if b.selection.start_row == b.selection.end_row:
                    if b.selection.end_col - b.selection.start_col == len(b.getLine(b.selection.start_row)):
                        # 单行且选中了整行，此时不校正显示位置匹配
                        line = b.getLine(b.selection.start_row)
                    else:
                        # 单行且选中了部分内容，此时校正显示位置匹配
                        line = self.consoleView.line_correction(b.getLine(b.selection.start_row))

                    start = max(0, b.selection.start_col)
                    end = min(len(line), b.selection.end_col)
                    line_plain = Session.PLAIN_TEXT_REGX.sub("", line).replace("\r", "").replace("\x00", "")
                    selection = line_plain[start:end]
                    self.app.clipboard.set_text(selection)
                    self.set_status(Settings.gettext("msg_copy", selection))
                    if self.current_session:
                        self.current_session.setVariable("%copy", selection)
                else:
                    # 多行只认行
                    lines = []
                    for row in range(b.selection.start_row, b.selection.end_row + 1):
                        line = b.getLine(row)
                        line_plain = Session.PLAIN_TEXT_REGX.sub("", line).replace("\r", "").replace("\x00", "")
                        lines.append(line_plain)
                    copy_text = "\n".join(lines)
                    self.app.clipboard.set_text(copy_text)
                    self.set_status(Settings.gettext("msg_copylines", b.selection.rows))

            else:
                # RAW模式，直接复制原始内容
                if b.selection.start_row == b.selection.end_row:
                    # 单行情况
                    line = b.getLine(b.selection.start_row)
                    self.app.clipboard.set_text(line)
                    self.set_status(Settings.gettext("msg_copy", line))
                    if self.current_session:
                        self.current_session.setVariable("%copy", line)

                else:
                    # 多行只认行
                    lines = []
                    for row in range(b.selection.start_row, b.selection.end_row + 1):
                        line = b.getLine(row)
                        lines.append(line)
                    copy_raw_text = "\n".join(lines)
                    self.app.clipboard.set_text(copy_raw_text)
                    self.set_status(Settings.gettext("msg_copylines", b.selection.rows))

                    if self.current_session:
                        self.current_session.setVariable("%copy", copy_raw_text)


        else:
            self.set_status(Settings.gettext("msg_no_selection"))

    def create_session(self, name, host, port, encoding = None, after_connect = None, scripts = None, userid = None):
        """
        创建一个会话。菜单或者#session命令均调用本函数执行创建会话。

        :param name: 会话名称
        :param host: 服务器域名或IP地址
        :param port: 端口号
        :param encoding: 服务器编码
        :param after_connect: 连接后要向服务器发送的内容，用来实现自动登录功能
        :param scripts: 要加载的脚本清单
        :param userid: 自动登录的ID(获取自cfg文件中的定义，绑定到菜单)，将以该值在该会话中创建一个名为id的变量
        """
        result = False
        encoding = encoding or Settings.server["default_encoding"]

        if name not in self.sessions.keys():
            session = Session(self, name, host, port, encoding, after_connect, scripts = scripts)
            session.setVariable("id", userid)
            self.sessions[name] = session
            self.activate_session(name)

            for plugin in self._plugins.values():
                if isinstance(plugin, Plugin):
                    plugin.onSessionCreate(session)

            result = True
        else:
            self.set_status(Settings.gettext("msg_session_exists", name))

        return result

    def show_logSelectDialog(self):
        def correction_align_width(text, width):
            "修正文本对齐宽度，防止ljust和rjust方法产生的中文宽度不对齐问题"
            return width - wcswidth(text) + len(text)
        async def coroutine():
            title_filename = Settings.gettext("logfile_name").ljust(correction_align_width(Settings.gettext("logfile_name"), 20))
            title_filesize = Settings.gettext("logfile_size").rjust(correction_align_width(Settings.gettext("logfile_size"), 20))
            title_modified = Settings.gettext("logfile_modified").center(correction_align_width(Settings.gettext("logfile_modified"), 23))
            head_line = "   {}{}{}".format(title_filename, title_filesize, title_modified)
            
            log_list = list()
            files = [f for f in os.listdir('.') if os.path.isfile(f) and f.endswith('.log')]
            for file in files:
                file = os.path.abspath(file)
                filename = os.path.basename(file).ljust(20)
                filesize = f"{os.path.getsize(file):,} Bytes".rjust(20)
                # ctime   = datetime.fromtimestamp(os.path.getctime(file)).strftime('%Y-%m-%d %H:%M:%S').rjust(23)
                mtime   = datetime.fromtimestamp(os.path.getmtime(file)).strftime('%Y-%m-%d %H:%M:%S').rjust(23)
                
                file_display_line = "{}{}{}".format(filename, filesize, mtime)
                log_list.append((file, file_display_line))

            logDir = os.path.abspath(os.path.join(os.curdir, 'log'))
            if os.path.exists(logDir):
                files = [f for f in os.listdir(logDir) if f.endswith('.log')]
                for file in files:
                    file = os.path.join(logDir, file)
                    filename = ('log/' + os.path.basename(file)).ljust(20)
                    filesize = f"{os.path.getsize(file):,} Bytes".rjust(20)
                    # ctime   = datetime.fromtimestamp(os.path.getctime(file)).strftime('%Y-%m-%d %H:%M:%S').rjust(23)
                    mtime   = datetime.fromtimestamp(os.path.getmtime(file)).strftime('%Y-%m-%d %H:%M:%S').rjust(23)
                    
                    file_display_line = "{}{}{}".format(filename, filesize, mtime)
                    log_list.append((file, file_display_line))
            
            dialog = LogSelectionDialog(
                text = head_line,
                values = log_list
            )

            result = await self.show_dialog_as_float(dialog)

            if result:
                self.logFileShown = result
                self.showLogInTab()

        asyncio.ensure_future(coroutine())

    def showLogInTab(self):
        "在记录也显示LOG记录"
        self.current_session = None
        self.showLog = True

        if self.logFileShown:
            filename = os.path.abspath(self.logFileShown)
            if os.path.exists(filename):
                lock = threading.RLock()
                lock.acquire()
                self.logSessionBuffer.loadfile(filename, encoding = 'utf-8', errors = 'ignore')
                lock.release()

            #self.logSessionBuffer.cursor_position = len(self.logSessionBuffer.text)
            self.consoleView.buffer = self.logSessionBuffer
            self.app.invalidate()

    def activate_session(self, key):
        "激活指定名称的session，并将该session设置为当前session"
        session = self.sessions.get(key, None)

        if isinstance(session, Session):
            self.current_session = session
            self.consoleView.buffer = session.buffer
            #self.set_status(Settings.text["session_changed"].format(session.name))
            self.app.invalidate()

    def close_session(self):
        "关闭当前会话。若当前会话处于连接状态，将弹出对话框以确认。"
        async def coroutine():
            if self.current_session:
                if self.current_session.connected:
                    dlgQuery = QueryDialog(HTML(f'<b fg="red">{Settings.gettext("warning")}</b>'), HTML(f'<style fg="red">{Settings.gettext("session_close_prompt", self.current_session.name)}</style>'))
                    result = await self.show_dialog_as_float(dlgQuery)
                    if result:
                        self.current_session.disconnect()

                        # 增加延时等待确保会话关闭
                        wait_time = 0
                        while self.current_session.connected:
                            await asyncio.sleep(0.1)
                            wait_time += 1
                            if wait_time > 100:
                                self.current_session.onDisconnected(None)
                                break
                            
                    else:
                        return

                for plugin in self._plugins.values():
                    if isinstance(plugin, Plugin):
                        plugin.onSessionDestroy(self.current_session)

                name = self.current_session.name
                self.current_session.closeLoggers()
                self.current_session.clean()
                self.current_session = None
                #self.consoleView.buffer = SessionBuffer()
                self.consoleView.buffer = None
                self.sessions.pop(name)
                #self.set_status(f"会话 {name} 已关闭")
                if len(self.sessions.keys()) > 0:
                    new_sess = list(self.sessions.keys())[0]
                    self.activate_session(new_sess)
                    #self.set_status(f"当前会话已切换为 {self.current_session.name}")

        asyncio.ensure_future(coroutine()) # type: ignore

    # 菜单选项操作 - 开始

    def act_new(self):
        "菜单: 创建新会话"
        async def coroutine():
            dlgNew = NewSessionDialog()
            result = await self.show_dialog_as_float(dlgNew)
            if result:
                self.create_session(*result)
            return result
        
        asyncio.ensure_future(coroutine())

    def act_connect(self):
        "菜单: 连接/重新连接"
        if self.current_session:
            self.current_session.handle_connect()

    def act_discon(self):
        "菜单: 断开连接"
        if self.current_session:
            self.current_session.disconnect()

    def act_nosplit(self):
        "菜单: 取消分屏"
        if self.current_session:
            s = self.current_session
            b = s.buffer
            b.exit_selection()
            #b.cursor_position = len(b.text)
            b.nosplit()

        elif self.showLog:
            b = self.logSessionBuffer
            b.exit_selection()
            #b.cursor_position = len(b.text)
            #b.start_lineno = -1
            b.nosplit()

    def act_close_session(self):
        "菜单: 关闭当前会话"
        if self.current_session:
            self.close_session()

        elif self.showLog:
            self.showLog = False
            #self.logSessionBuffer = None
            if len(self.sessions.keys()) > 0:
                new_sess = list(self.sessions.keys())[0]
                self.activate_session(new_sess)

    def act_beautify(self):
        "菜单: 打开/关闭美化显示"
        val = not Settings.client["beautify"]
        Settings.client["beautify"] = val
        if self.current_session:
            self.current_session.info(f'{Settings.gettext("msg_beautify")}{Settings.gettext("msg_open") if val else Settings.gettext("msg_close")}!')

    def act_echoinput(self):
        "菜单: 显示/隐藏输入指令"
        val = not Settings.client["echo_input"]
        Settings.client["echo_input"] = val
        if self.current_session:
            self.current_session.info(f'{Settings.gettext("msg_echoinput")}{Settings.gettext("msg_open") if val else Settings.gettext("msg_close")}!')

    def act_autoreconnect(self):
        "菜单: 打开/关闭自动重连"
        val = not Settings.client["auto_reconnect"]
        Settings.client["auto_reconnect"] = val
        if self.current_session:
            self.current_session.info(f'{Settings.gettext("msg_autoreconnect")}{Settings.gettext("msg_open") if val else Settings.gettext("msg_close")}')

    def act_copy(self):
        "菜单: 复制纯文本"
        self.copy()

    def act_copyraw(self):
        "菜单: 复制(ANSI)"
        self.copy(raw = True)

    def act_clearsession(self):
        "菜单: 清空会话内容"
        if self.consoleView.buffer:
            self.consoleView.buffer.clear()

    def act_reload(self):
        "菜单: 重新加载脚本配置"
        if self.current_session:
            self.current_session.handle_reload()

    # 暂未实现该功能
    def act_change_layout(self, layout):
        self.status_display = layout
        self.app.invalidate()

    def act_exit(self):
        """菜单: 退出"""
        async def coroutine():
            con_sessions = list()
            for session in self.sessions.values():
                if session.connected:
                    con_sessions.append(session.name)

            if len(con_sessions) > 0:
                dlgQuery = QueryDialog(HTML(f'<b fg="red">{Settings.gettext("warning_exit")}</b>'), HTML(f'<style fg="red">{Settings.gettext("app_exit_prompt", len(con_sessions), ", ".join(con_sessions))}</style>'))
                result = await self.show_dialog_as_float(dlgQuery)
                if result:
                    for ss_name in con_sessions:
                        ss = self.sessions[ss_name]
                        ss.disconnect()

                        # 增加延时等待确保会话关闭
                        wait_time = 0
                        while ss.connected:
                            await asyncio.sleep(0.1)
                            wait_time += 1
                            if wait_time > 100:
                                ss.onDisconnected(None)
                                break

                        for plugin in self._plugins.values():
                            if isinstance(plugin, Plugin):
                                plugin.onSessionDestroy(ss)

                else:
                    return
                
            self.app.exit()

        asyncio.ensure_future(coroutine())

    def act_about(self):
        "菜单: 关于"
        dialog_about = WelcomeDialog(True)
        self.show_dialog(dialog_about)

    # 菜单选项操作 - 完成

    def get_input_prompt(self):
        "命令输入行提示符"
        return HTML(Settings.gettext("input_prompt"))

    def btn_title_clicked(self, name, mouse_event: MouseEvent):
        "顶部会话标签点击切换鼠标事件"
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            if name == '[LOG]':
                self.showLogInTab()
            else:
                self.activate_session(name)

    def get_frame_title(self):
        "顶部会话标题选项卡"
        if len(self.sessions.keys()) == 0:
            if not self.showLog:
                return Settings.__appname__ + " " + Settings.__version__
            else:
                if self.logFileShown:
                    return f'[LOG] {self.logFileShown}'
                else:
                    return f'[LOG]'
        
        title_formatted_list = []
        for key, session in self.sessions.items():
            if session == self.current_session:
                if session.connected:
                    style = Settings.styles["selected.connected"]
                else:
                    style = Settings.styles["selected"]

            else:
                if session.connected:
                    style = Settings.styles["normal.connected"]
                else:
                    style = Settings.styles["normal"]

            title_formatted_list.append((style, key, functools.partial(self.btn_title_clicked, key)))
            title_formatted_list.append(("", " | "))

        if self.showLog:
            if self.current_session is None:
                style = style = Settings.styles["selected"]
            else:
                style = Settings.styles["normal"]

            title = f'[LOG] {self.logFileShown}' if self.logFileShown else f'[LOG]'

            title_formatted_list.append((style, title, functools.partial(self.btn_title_clicked, '[LOG]')))
            title_formatted_list.append(("", " | "))

        return title_formatted_list[:-1]

    def get_statusbar_text(self):
        "状态栏内容"
        return [
            ("class:status", " "),
            ("class:status", self.status_message),
        ]
    
    def get_statusbar_right_text(self):
        "状态栏右侧内容"
        con_str, mouse_support, tri_status, beautify = "", "", "", ""
        if not Settings.client["beautify"]:
            beautify = Settings.gettext("status_nobeautify") + " "

        if not self._mouse_support:
            mouse_support = Settings.gettext("status_mouseinh") + " "

        if self.current_session:
            if self.current_session._ignore:
                tri_status = Settings.gettext("status_ignore") + " "

            if not self.current_session.connected:
                con_str = Settings.gettext("status_notconnect")
            else:
                dura = self.current_session.duration
                DAY, HOUR, MINUTE = 86400, 3600, 60
                days, hours, mins, secs = 0,0,0,0
                days = dura // DAY
                dura = dura - days * DAY
                hours = dura // HOUR
                dura = dura - hours * HOUR
                mins = dura // MINUTE
                sec = dura - mins * MINUTE
                if days > 0:
                    con_str = Settings.gettext("status_connected") + ": {0:.0f}{4}{1:.0f}{5}{2:.0f}{6}{3:.0f}{7}".format(days, hours, mins, sec, Settings.gettext("Day"), Settings.gettext("Hour"), Settings.gettext("Minute"), Settings.gettext("Second"))
                elif hours > 0:
                    con_str = Settings.gettext("status_connected") + ": {0:.0f}{3}{1:.0f}{4}{2:.0f}{5}".format(hours, mins, sec, Settings.gettext("Hour"), Settings.gettext("Minute"), Settings.gettext("Second"))
                elif mins > 0:
                    con_str = Settings.gettext("status_connected") + ": {0:.0f}{2}{1:.0f}{3}".format(mins, sec, Settings.gettext("Minute"), Settings.gettext("Second"))
                else:
                    con_str = Settings.gettext("status_connected") + ": {:.0f}{}".format(sec, Settings.gettext("Second"))

        return "{}{}{}{} {} {} ".format(beautify, mouse_support, tri_status, con_str, Settings.__appname__, Settings.__version__)

    def get_statuswindow_text(self):
        "状态窗口: status_maker 的内容"
        text = ""

        try:
            if self.current_session:
                text = self.current_session.get_status()
        except Exception as e:
            text = f"{e}"

        return text

    def set_status(self, msg):
        """
        在状态栏中上显示消息。可在代码中调用
        
        :param msg: 要显示的消息
        """
        self.status_message = msg
        self.app.invalidate()

    def _quickHandleSession(self, group, name):
        '''
        根据指定的组名和会话角色名，从Settings内容，创建一个会话
        '''
        handled = False
        if name in self.sessions.keys():
           self.activate_session(name)
           handled = True

        else:
            site = Settings.sessions[group]
            if name in site["chars"].keys():
                host = site["host"]
                port = site["port"]
                encoding = site["encoding"]
                autologin = site["autologin"]
                default_script = site["default_script"]
                
                def_scripts = list()
                if isinstance(default_script, str):
                    def_scripts.extend(default_script.split(","))
                elif isinstance(default_script, (list, tuple)):
                    def_scripts.extend(default_script)

                charinfo = site["chars"][name]

                after_connect = autologin.format(charinfo[0], charinfo[1])
                sess_scripts = list()
                sess_scripts.extend(def_scripts)
                    
                if len(charinfo) == 3:
                    session_script = charinfo[2]
                    if session_script:
                        if isinstance(session_script, str):
                            sess_scripts.extend(session_script.split(","))
                        elif isinstance(session_script, (list, tuple)):
                            sess_scripts.extend(session_script)

                self.create_session(name, host, port, encoding, after_connect, sess_scripts, charinfo[0])
                handled = True
        
        return handled


    def handle_session(self, *args):
        '''
        嵌入命令 #session 的执行函数，创建一个远程连接会话。
        该函数不应该在代码中直接调用。

        使用:
            - #session {name} {host} {port} {encoding}
            - 当不指定 Encoding: 时, 默认使用utf-8编码
            - 可以直接使用 #{名称} 切换会话和操作会话命令

            - #session {group}.{name}
            - 相当于直接点击菜单{group}下的{name}菜单来创建会话. 当该会话已存在时，切换到该会话

        参数:
            :name: 会话名称
            :host: 服务器域名或IP地址
            :port: 端口号
            :encoding: 编码格式，不指定时默认为 utf8
    
            :group: 组名, 即配置文件中, sessions 字段下的某个关键字
            :name: 会话快捷名称, 上述 group 关键字下的 chars 字段中的某个关键字

        示例:
            ``#session {名称} {宿主机} {端口} {编码}`` 
                创建一个远程连接会话，使用指定编码格式连接到远程宿主机的指定端口并保存为 {名称} 。其中，编码可以省略，此时使用Settings.server["default_encoding"]的值，默认为utf8
            ``#session newstart mud.pkuxkx.net 8080 GBK`` 
                使用GBK编码连接到mud.pkuxkx.net的8080端口，并将该会话命名为newstart
            ``#session newstart mud.pkuxkx.net 8081`` 
                使用UTF8编码连接到mud.pkuxkx.net的8081端口，并将该会话命名为newstart
            ``#newstart`` 
                将名称为newstart的会话切换为当前会话
            ``#newstart give miui gold`` 
                使名称为newstart的会话执行give miui gold指令，但不切换到该会话

            ``#session pkuxkx.newstart``
                通过指定快捷配置创建会话，相当于点击 世界->pkuxkx->newstart 菜单创建会话。若该会话存在，则切换到该会话

        相关命令:
            - #close
            - #exit

        '''

        nothandle = True
        errmsg = "错误的#session命令"
        if len(args) == 1:
            host_session = args[0]
            if '.' in host_session:
                group, name = host_session.split('.')
                nothandle = not self._quickHandleSession(group, name)

            else:
                errmsg = Settings.gettext("msg_cmd_session_error")

        elif len(args) >= 3:
            session_name = args[0]
            session_host = args[1]
            session_port = int(args[2])
            if len(args) == 4:
                session_encoding = args[3]
            else:
                session_encoding = Settings.server["default_encoding"]

            self.create_session(session_name, session_host, session_port, session_encoding)
            nothandle = False
        
        if nothandle:
            self.set_status(errmsg)

    def enter_pressed(self, buffer: Buffer):
        "命令行回车按键处理"
        cmd_line = buffer.text
        space_index = cmd_line.find(" ")
        
        if len(cmd_line) == 0:
            if self.current_session:
                self.current_session.writeline("")
        
        elif cmd_line[0] != Settings.client["appcmdflag"]:
            if self.current_session:
                self.current_session.last_command = cmd_line

        if cmd_line.startswith("#session"):
            cmd_tuple = cmd_line[1:].split()
            self.handle_session(*cmd_tuple[1:])

        else:
            if self.current_session:
                if len(cmd_line) == 0:
                    self.current_session.writeline("")
                else:
                    try:
                        self.current_session.log.log(f"{Settings.gettext('msg_cmdline_input')} {cmd_line}\n")

                        cb = CodeBlock(cmd_line)
                        cb.execute(self.current_session)
                    except Exception as e:
                        self.current_session.warning(e)
                        self.current_session.exec_command(cmd_line)
            else:
                if cmd_line == "#exit":
                    self.act_exit()
                elif (cmd_line == "#close") and self.showLog:
                    self.act_close_session()
                else:
                    self.set_status(Settings.gettext("msg_no_session"))

        # 配置：命令行内容保留
        if Settings.client["remain_last_input"]:
            buffer.cursor_position = 0
            buffer.start_selection()
            buffer.cursor_right(len(cmd_line))
            return True
        
        else:
            return False

    @property
    def globals(self):
        """
        全局变量，快捷点访问器
        用于替代get_globals与set_globals函数的调用
        """
        return self._globals

    def get_globals(self, name, default = None):
        """
        获取PYMUD全局变量
        
        :param name: 全局变量名称
        :param default: 当全局变量不存在时的返回值
        """
        if name in self._globals.keys():
            return self._globals[name]
        else:
            return default

    def set_globals(self, name, value):
        """
        设置PYMUD全局变量

        :param name: 全局变量名称
        :param value: 全局变量值。值可以为任何类型。
        """
        self._globals[name] = value

    def del_globals(self, name):
        """
        移除一个PYMUD全局变量
        移除全局变量是从字典中删除该变量，而不是将其设置为None

        :param name: 全局变量名称
        """
        if name in self._globals.keys():
            self._globals.pop(name)

    @property
    def plugins(self):
        "所有已加载的插件列表，快捷点访问器"
        return self._plugins

    def show_message(self, title, text, modal = True):
        "显示一个消息对话框"
        async def coroutine():
            dialog = MessageDialog(title, text, modal)
            await self.show_dialog_as_float(dialog)

        asyncio.ensure_future(coroutine())

    def show_dialog(self, dialog):
        "显示一个给定的对话框"
        async def coroutine():
            await self.show_dialog_as_float(dialog)

        asyncio.ensure_future(coroutine())

    async def show_dialog_as_float(self, dialog):
        "显示弹出式窗口."
        float_ = Float(content=dialog)
        if self.root_container.floats:

            self.root_container.floats.insert(0, float_)

            self.app.layout.focus(dialog)
            result = await dialog.future
            self.app.layout.focus(self.commandLine)

            if float_ in self.root_container.floats:
                self.root_container.floats.remove(float_)

            return result

    async def run_async(self):
        "以异步方式运行本程序"
        # 运行插件启动应用，放在此处，确保插件初始化在event_loop创建完成之后运行
        for plugin in self._plugins.values():
            if isinstance(plugin, Plugin):
                plugin.onAppInit(self)
                
        asyncio.create_task(self.onSystemTimerTick())
        await self.app.run_async(set_exception_handler = False)

        # 当应用退出时，运行插件销毁应用
        for plugin in self._plugins.values():
            if isinstance(plugin, Plugin):
                plugin.onAppDestroy(self)

    def run(self):
        "运行本程序"
        #self.app.run(set_exception_handler = False)
        asyncio.run(self.run_async())

    def get_width(self):
        "获取ConsoleView的实际宽度，等于输出宽度,（已经没有左右线条和滚动条了）"
        size = self.app.output.get_size().columns
        if Settings.client["status_display"] == 2:
            size = size - Settings.client["status_width"] - 1
        return size

    def get_height(self):
        "获取ConsoleView的实际高度，等于输出高度-5,（上下线条，菜单，命令栏，状态栏）"
        size = self.app.output.get_size().rows - 5

        if Settings.client["status_display"] == 1:
            size = size - Settings.client["status_height"] - 1
        return size

    #####################################
    # plugins 处理
    #####################################
    def load_plugins(self):
        "加载插件。将加载pymud包的plugins目录下插件，以及当前目录的plugins目录下插件"
        # 首先加载系统目录下的插件
        current_dir = os.path.dirname(__file__)
        plugins_dir = os.path.join(current_dir, "plugins")
        if os.path.exists(plugins_dir):
            for file in os.listdir(plugins_dir):
                if file.endswith(".py"):
                    try:
                        file_path = os.path.join(plugins_dir, file)
                        file_name = file[:-3]
                        plugin = Plugin(file_name, file_path)
                        self._plugins[plugin.name] = plugin
                        # plugin.onAppInit(self)
                    except Exception as e:
                        self.set_status(Settings.gettext("msg_plugin_load_error", file, e))
        
        # 然后加载当前目录下的插件
        current_dir = os.path.abspath(".")
        plugins_dir = os.path.join(current_dir, "plugins")
        if os.path.exists(plugins_dir):
            for file in os.listdir(plugins_dir):
                if file.endswith(".py"):
                    try:
                        file_path = os.path.join(plugins_dir, file)
                        file_name = file[:-3]
                        plugin = Plugin(file_name, file_path)
                        self._plugins[plugin.name] = plugin
                        plugin.onAppInit(self)
                    except Exception as e:
                        self.set_status(Settings.gettext("msg_plugin_load_error", file, e))

    def reload_plugin(self, plugin: Plugin):
        "重新加载指定插件"
        for session in self.sessions.values():
            plugin.onSessionDestroy(session)

        plugin.reload()
        plugin.onAppInit(self)

        for session in self.sessions.values():
            plugin.onSessionCreate(session)
        

def startApp(cfg_data = None):
    app = PyMudApp(cfg_data)
    app.run()



import sys, functools, traceback
from typing import Union, Optional, List

def print_exception(session, e: Exception):
    """打印异常信息"""
    from .settings import Settings
    from .session import Session
    if isinstance(session, Session):
        # tb = sys.exc_info()[2]
        # frames = traceback.extract_tb(tb)
        # frame = frames[-1]
        # session.error(Settings.gettext("exception_traceback", frame.filename, frame.lineno, frame.name), Settings.gettext("script_error"))
        # if frame.line:
        #     session.error(f"    {frame.line}", Settings.gettext("script_error"))

        # session.error(Settings.gettext("exception_message", type(e).__name__, e), Settings.gettext("script_error"))
        # session.error("===========================", Settings.gettext("script_error"))
        session.error(traceback.format_exc(), Settings.gettext("script_error"))

def exception(func):
    """方法异常处理装饰器，捕获异常后通过会话的session.error打印相关信息"""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        from .objects import BaseObject
        from .modules import ModuleInfo, IConfig
        from .session import Session
        from .settings import Settings
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            # 调用类的错误处理方法
            if isinstance(self, Session):
                session = self
            elif isinstance(self, (BaseObject, IConfig, ModuleInfo)):
                session = self.session
            else:
                session = None
                
            if isinstance(session, Session):
                print_exception(session, e)
                #session.error(Settings.gettext("exception_message", e, type(e)))
                #session.error(Settings.gettext("exception_traceback", traceback.format_exc()))
            else:
                raise  # 当没有会话时，选择重新抛出异常
    return wrapper

def async_exception(func):
    """异步方法异常处理装饰器，捕获异常后通过会话的session.error打印相关信息"""
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        from .objects import BaseObject
        from .modules import ModuleInfo, IConfig
        from .session import Session
        from .settings import Settings
        try:
            return await func(self, *args, **kwargs)
        except Exception as e:
            if isinstance(self, Session):
                session = self
            elif isinstance(self, (BaseObject, IConfig, ModuleInfo)):
                session = self.session
            else:
                session = None

            if isinstance(session, Session):
                print_exception(session, e)

            else:
                raise  # 当没有会话时，选择重新抛出异常
    return wrapper


class PymudDecorator:
    """
    装饰器基类。使用装饰器可以快捷创建各类Pymud基础对象。
    
        :param type: 装饰器的类型，用于区分不同的装饰器，为字符串类型。
        :param args: 可变位置参数，用于传递额外的参数。
        :param kwargs: 可变关键字参数，用于传递额外的键值对参数。
    """
    def __init__(self, type: str, *args, **kwargs):
        self.type = type
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        decos = getattr(func, "__pymud_decorators__", [])
        decos.append(self)
        func.__pymud_decorators__ = decos

        return func

    def __repr__(self):
        return f"<{self.__class__.__name__} type = {self.type} args = {self.args} kwargs = {self.kwargs}>"


def alias(
        patterns: Union[List[str], str], 
        id: Optional[str] = None, 
        group: str = "", 
        enabled: bool = True, 
        ignoreCase: bool = False, 
        isRegExp: bool = True, 
        keepEval: bool = False, 
        expandVar: bool = True, 
        priority: int = 100, 
        oneShot: bool = False,  
        *args, **kwargs):
    """
    使用装饰器创建一个别名（Alias）对象。被装饰的函数将在别名成功匹配时调用。
    被装饰的函数，除第一个参数为类实例本身self之外，另外包括id, line, wildcards三个参数。
    其中id为别名对象的唯一标识符，line为匹配的文本行，wildcards为匹配的通配符列表。

    :param patterns: 别名匹配的模式。
    :param id: 别名对象的唯一标识符，不指定时将自动生成唯一标识符。
    :param group: 别名所属的组名，默认为空字符串。
    :param enabled: 别名是否启用，默认为 True。
    :param ignoreCase: 匹配时是否忽略大小写，默认为 False。
    :param isRegExp: 模式是否为正则表达式，默认为 True。
    :param keepEval: 若存在多个可匹配的别名时，是否持续匹配，默认为 False。
    :param expandVar: 是否展开变量，默认为 True。
    :param priority: 别名的优先级，值越小优先级越高，默认为 100。
    :param oneShot: 别名是否只触发一次后自动失效，默认为 False。
    :param args: 可变位置参数，用于传递额外的参数。
    :param kwargs: 可变关键字参数，用于传递额外的键值对参数。
    :return: PymudDecorator 实例，类型为 "alias"。
    """
    # 将传入的参数更新到 kwargs 字典中
    kwargs.update({
        "patterns": patterns,
        "id": id, 
        "group": group, 
        "enabled": enabled, 
        "ignoreCase": ignoreCase, 
        "isRegExp": isRegExp, 
        "keepEval": keepEval, 
        "expandVar": expandVar, 
        "priority": priority, 
        "oneShot": oneShot})
    
    # 如果 id 为 None，则从 kwargs 中移除 "id" 键
    if not id:
        kwargs.pop("id")

    return PymudDecorator("alias", *args, **kwargs)

def trigger(
        patterns: Union[List[str], str], 
        id: Optional[str] = None, 
        group: str = "", 
        enabled: bool = True, 
        ignoreCase: bool = False, 
        isRegExp: bool = True, 
        keepEval: bool = False, 
        expandVar: bool = True, 
        raw: bool = False, 
        priority: int = 100, 
        oneShot: bool = False,  
        *args, **kwargs):
    """
    使用装饰器创建一个触发器（Trigger）对象。

    :param patterns: 触发器匹配的模式。单行模式可以是字符串或正则表达式，多行模式必须是元组或列表，其中每个元素都是字符串或正则表达式。
    :param id: 触发器对象的唯一标识符，不指定时将自动生成唯一标识符。
    :param group: 触发器所属的组名，默认为空字符串。
    :param enabled: 触发器是否启用，默认为 True。
    :param ignoreCase: 匹配时是否忽略大小写，默认为 False。
    :param isRegExp: 模式是否为正则表达式，默认为 True。
    :param keepEval: 若存在多个可匹配的触发器时，是否持续匹配，默认为 False。
    :param expandVar: 是否展开变量，默认为 True。
    :param raw: 是否使用原始匹配方式，默认为 False。原始匹配方式下，不对 VT100 下的 ANSI 颜色进行解码，因此可以匹配颜色；正常匹配仅匹配文本。
    :param priority: 触发器的优先级，值越小优先级越高，默认为 100。
    :param oneShot: 触发器是否只触发一次后自动失效，默认为 False。
    :param args: 可变位置参数，用于传递额外的参数。
    :param kwargs: 可变关键字参数，用于传递额外的键值对参数。
    :return: PymudDecorator 实例，类型为 "trigger"。
    """
    # 将传入的参数更新到 kwargs 字典中
    kwargs.update({
        "patterns": patterns,
        "id": id, 
        "group": group, 
        "enabled": enabled, 
        "ignoreCase": ignoreCase, 
        "isRegExp": isRegExp, 
        "keepEval": keepEval, 
        "expandVar": expandVar, 
        "raw": raw,
        "priority": priority, 
        "oneShot": oneShot})
    if not id:
        kwargs.pop("id")
    return PymudDecorator("trigger", *args, **kwargs)

def timer(timeout: float, id: Optional[str] = None, group: str = "", enabled: bool = True, *args, **kwargs):
    """
    使用装饰器创建一个定时器（Timer）对象。

    :param timeout: 定时器超时时间，单位为秒。
    :param id: 定时器对象的唯一标识符，不指定时将自动生成唯一标识符。
    :param group: 定时器所属的组名，默认为空字符串。
    :param enabled: 定时器是否启用，默认为 True。
    :param args: 可变位置参数，用于传递额外的参数。
    :param kwargs: 可变关键字参数，用于传递额外的键值对参数。
    :return: PymudDecorator 实例，类型为 "timer"。
    """
    kwargs.update({
        "timeout": timeout,
        "id": id,
        "group": group,
        "enabled": enabled
        })
    if not id:
        kwargs.pop("id")
    return PymudDecorator("timer", *args, **kwargs)

def gmcp(name: str, group: str = "", enabled: bool = True, *args, **kwargs):
    """
    使用装饰器创建一个GMCP触发器（GMCPTrigger）对象。

    :param name: GMCP触发器的名称。
    :param group: GMCP触发器所属的组名，默认为空字符串。
    :param enabled: GMCP触发器是否启用，默认为 True。
    :param args: 可变位置参数，用于传递额外的参数。
    :param kwargs: 可变关键字参数，用于传递额外的键值对参数。
    :return: PymudDecorator 实例，类型为 "gmcp"。
    """
    kwargs.update({
        "id": name,
        "group": group,
        "enabled": enabled
        })
    
    return PymudDecorator("gmcp", *args, **kwargs)
